static char rcsid[] = "@(#)$Id: sb_mem.c,v 2.4 2013-02-09 06:55:21 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.4 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "s_me.h"
#include "sb_imp.h"
#include "pg_range_imp.h"

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

#define SB_in_mem_magic	     0xEB08

struct sb_in_mem {
    unsigned short magic;     /* SB_in_mem_magic */

    int              string_count;
    struct sb_line_data {
	struct string *line;
	int            pg_flags;
	struct pager_range *range;
    } * strings;

    /* Not really used a*/
    struct sb_pager_range {
	struct pgrng_bfr_binding *pager_range;
    } * pager_range;
    int pager_range_count;
    
};


S_(sb_init_buffer sb_init_mem)
static int sb_init_mem P_((struct stringbuffer *buffer));
static int sb_init_mem(buffer)
     struct stringbuffer *buffer;
{
    buffer->a.mem = safe_malloc(sizeof (* (buffer->a.mem)));

    /* defined in hdrs/defs.h */
    bzero((void *)buffer->a.mem,sizeof (* (buffer->a.mem)));

    buffer->a.mem->magic = SB_in_mem_magic;

    buffer->a.mem->string_count = 0;
    buffer->a.mem->strings      = NULL;

    buffer->a.mem->pager_range = NULL;
    buffer->a.mem->pager_range_count = 0;

    return 1;
}

S_(sb_free_buffer sb_free_mem)
static void sb_free_mem P_((struct stringbuffer *buffer));
static void sb_free_mem(buffer)
     struct stringbuffer *buffer;
{
    if (SB_in_mem_magic != buffer->a.mem->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_free_mem",
	      "Bad magic number",0);
        
    if (buffer->a.mem->strings) {
	int i;

	for (i = 0; i < buffer->a.mem->string_count; i++) {
	    if (buffer->a.mem->strings[i].line)
		free_string(&(buffer->a.mem->strings[i].line));
	    if (buffer->a.mem->strings[i].range)
		free_pager_range(&(buffer->a.mem->strings[i].range));
	}
	free(buffer->a.mem->strings);
	buffer->a.mem->strings      = NULL;
    }
    buffer->a.mem->string_count = 0;

    if (buffer->a.mem->pager_range) {
	int i;

	for (i = 0; i < buffer->a.mem->pager_range_count; i++) {
	    if (buffer->a.mem->pager_range[i].pager_range)
		pgrng_bfr_bnd_clear_stringbuffer(& (buffer->a.mem->
						    pager_range[i].pager_range),
						 buffer,i);
	}
    }
    buffer->a.mem->pager_range_count = 0;
}

S_(sb_add_line_to_stringbuffer sb_add_line_to_mem)
static void sb_add_line_to_mem P_((struct stringbuffer *buffer,
				   const struct string *string,
				   int pg_flags,struct pager_range *range,
				   struct pgrng_bfr_binding  * binding));
static void sb_add_line_to_mem(buffer,string,pg_flags,range,binding)
     struct stringbuffer *buffer;
     const struct string *string;
     int pg_flags;
     struct pager_range *range;    
     struct pgrng_bfr_binding  * binding;  /* NOT USED */
{
    int ptr;

    if (SB_in_mem_magic != buffer->a.mem->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_add_line_to_mem",
	      "Bad magic number",0);

    ptr = buffer->a.mem->string_count;
    
    buffer->a.mem->strings = safe_realloc(buffer->a.mem->strings,
					    (buffer->a.mem->string_count+1) *
					    sizeof (buffer->a.mem->strings[0]));

    buffer->a.mem->strings[ptr].line     = dup_string(string);
    buffer->a.mem->strings[ptr].pg_flags = pg_flags;
    buffer->a.mem->strings[ptr].range     = range;
    if (buffer->a.mem->strings[ptr].range)
	inc_pager_range_refcount(buffer->a.mem->strings[ptr].range);

    buffer->a.mem->string_count++;
}

S_(sb_linecount_stringbuffer sb_linecount_mem)
static int sb_linecount_mem P_((const struct stringbuffer *ptr));
static int sb_linecount_mem(ptr)
     const struct stringbuffer *ptr;
{
    if (SB_in_mem_magic != ptr->a.mem->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_linecount_mem",
	      "Bad magic number",0);

    return ptr->a.mem->string_count;
}

S_(sb_get_line_from_stringbuffer sb_get_line_from_mem)
static struct string *sb_get_line_from_mem P_((struct stringbuffer *buffer,
					       int ptr, int *pg_flags,
					       struct pager_range **range));
static struct string *sb_get_line_from_mem(buffer,ptr,pg_flags,range)
     struct stringbuffer *buffer;
     int ptr;
     int *pg_flags;
     struct pager_range **range;  
{
    if (SB_in_mem_magic != buffer->a.mem->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_mem",
	      "Bad magic number",0);

    if (ptr < 0 || ptr >= buffer->a.mem->string_count)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,"sb_get_line_from_mem",
	      "Bad index",0);

    if (range &&
	*range != buffer->a.mem->strings[ptr].range) {
	
	if (*range)
	    free_pager_range(range);

	*range = buffer->a.mem->strings[ptr].range;
	if (*range)
	    inc_pager_range_refcount(*range);	    
    }

    *pg_flags = buffer->a.mem->strings[ptr].pg_flags;
    return dup_string(buffer->a.mem->strings[ptr].line);

}

S_(sb_add_pager_range_to_stringbuffer sb_add_pager_range_to_mem)
static void sb_add_pager_range_to_mem P_((struct stringbuffer * buffer,
					   struct pgrng_bfr_binding * binding,
					  struct pager_range * range));
static void sb_add_pager_range_to_mem(buffer,binding,range)
     struct stringbuffer * buffer;
     struct pgrng_bfr_binding * binding;
     struct pager_range * range;
{
    int ptr;

    if (SB_in_mem_magic != buffer->a.mem->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "sb_add_pager_range_to_mem",
	      "Bad magic number",0);

    /* Because index numbers are not required,
       also this cache is not needed really */

    ptr = buffer->a.mem->pager_range_count;

    buffer->a.mem->pager_range = 
	safe_realloc(buffer->a.mem->pager_range,
		     (buffer->a.mem->pager_range_count +1) *
		     sizeof (buffer->a.mem->pager_range[0]));
    buffer->a.mem->pager_range[ptr].pager_range = binding;
    pgrng_bfr_bnd_set_stringbuffer(buffer->a.mem->pager_range[ptr].
				   pager_range,
				   buffer,ptr);
    buffer->a.mem->pager_range_count++;

}

S_(sb_get_pager_range_from_stringbuffer sb_get_pager_range_from_mem)
static struct pager_range * sb_get_pager_range_from_mem
  P_((struct stringbuffer * buffer,int index));
static struct pager_range * sb_get_pager_range_from_mem(buffer,index)
     struct stringbuffer * buffer;
     int index;
{
    struct pager_range *ret = NULL;
    /* Not used */

    if (SB_in_mem_magic != buffer->a.mem->magic)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "sb_get_pager_range_from_mem",
	      "Bad magic number",0);

    if (index < 0 || index >= buffer->a.mem->pager_range_count)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "sb_get_pager_range_from_mem",
	      "Bad index",0);
	
    if (! buffer->a.mem->pager_range[index].pager_range)
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "sb_get_pager_range_from_mem",
	      "No binding",0);

    ret = pgrng_bfr_bnd_get_range(buffer->a.mem->pager_range[index].pager_range);

    if (!ret) {
	panic("STRINGBUFFER PANIC",__FILE__,__LINE__,
	      "sb_get_pager_range_from_mem",
	      "pager range deallocated?",0);
    }
    
    return ret;
}

struct sb_type   sb_in_mem = {
    SB_TYPE_magic,

    sb_init_mem,
    sb_free_mem,
    sb_add_line_to_mem,
    sb_linecount_mem,
    sb_get_line_from_mem,
    sb_add_pager_range_to_mem,
    sb_get_pager_range_from_mem
};

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */

